// Copyright (C) 2018-2020  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package main
// XXX -> Package xbtree complements package lab.nexedi.com/kirr/neo/go/zodb/btree.
// TODO move -> btree when ΔTail matures.
// XXX doc

// FIXME the algorythm is different: recursion is implemented by expanding rangeSplit step by step.
//
// δ(BTree) notes
// ==============
//
// input: BTree, (@new, []oid)  ->  find out δ(BTree) i.e. {-k(v), +k'(v'), ...}
//
// - oid ∈ Bucket
// - oid ∈ BTree
//
// Bucket:
//
//      old = {k  -> v}
//      new = {k' -> v'}
//
//      Δ = -k(v), +k(v), ...
//
// => for all buckets
//
//      Δ accumulates to []δk(v)[n+,n-]  n+ ∈ {0,1}, n- ∈ {0,1}, if n+=n- - cancel
//
//
// BTree:
//
//      old = {k  -> B}   or {k  -> T}
//      new = {k' -> B'}  or {k' -> T'}
//
//      Δ = -k(B), +k(B), -k(T), +K(T), ...
//
// we translate (in top-down order):
//
//      k(B) -> {} of k(v)
//      k(T) -> {} of k(B) -> {} of k(v)
//
// which gives
//
//      Δ = k(v), +k(v), ...
//
// i.e. exactly as for buckets and it accumulates to global Δ.
//
// The globally-accumulated Δ is the answer for δ(BTree, (@new, []oid))
//
// top-down order is obtained via toposort({oid}) wrt visited PathSet.
//
// δ(BTree) in wcfs context:
//
// . -k(blk) -> invalidate #blk
// . +k(blk) -> invalidate #blk (e.g. if blk was previously read as hole)


//go:generate ./gen-set main Oid   Oid zset_oid.go

import (
	"context"
	"fmt"
	"math"
	"reflect"
	"sort"
	"strings"

	"lab.nexedi.com/kirr/go123/xerr"
	"lab.nexedi.com/kirr/neo/go/transaction"
	"lab.nexedi.com/kirr/neo/go/zodb"
	"lab.nexedi.com/kirr/neo/go/zodb/btree"
)

type Tree   = btree.LOBTree
type Bucket = btree.LOBucket
type Node   = btree.LONode

type TreeEntry   = btree.LOEntry
type BucketEntry = btree.LOBucketEntry

type Key    = int64
const KeyMax Key = math.MaxInt64
const KeyMin Key = math.MinInt64
type Value  = zodb.Oid // assumes key points to IPersistent
                       // deletion is represented as VDEL
const VDEL  = zodb.InvalidOid

type Oid    = zodb.Oid
type SetKey = SetI64

// ΔValue represents change in value.
type ΔValue struct {
	Old Value
	New Value
}


// ΔBtail represents tail of revisional changes to BTrees.
//
// It semantically consists of
//
//	[]δB			; rev ∈ (tail, head]
//	atTail
//
// where δB represents a change in BTrees space
//
//	δB:
//		.rev↑
//		{} root -> {}(key, value)
//
// and atTail keeps set of k/v @tail for keys changed in (tail, head]
//
//	atTail:
//		{} root -> {}(key, value)
//
// It covers only changes to keys from tracked subset of BTrees parts.
// In particular a key that was not explicitly requested to be tracked, even if
// it was changed in δZ, is not guaranted to be present in δB.
//
// ΔBtail provides the following operations:
//
//   .Track(path)		- start tracking tree nodes and keys; root=path[0], keys=path[-1].keys  XXX keys not correct - e.g. track missing key
//
//   .Update(δZ) -> δB				- update BTree δ tail given raw ZODB changes
//   .ForgetPast(revCut)			- forget changes past revCut
//   .SliceByRev(lo, hi) -> []δB		- XXX
//   .SliceByRootRev(root, lo, hi) -> []δT	- XXX
//   .Get(root, key, at)			- get root[key] @at assuming root[key] ∈ tracked
//
// An example for tracked set is a set of visited BTree paths.
// There is no requirement that tracked set belongs to only one single BTree.
//
// XXX δB can convert δZ to btree changes, but only at least for δZ's objects
// that ∈ BTree subgraphs that were explicitly requested to be tracked by δB.
//
// XXX incremental; not full coverage
// XXX see also zodb.ΔTail .
//
// ΔBtail is not safe for concurrent access.
// XXX -> multiple readers / single writer?
type ΔBtail struct {
	// raw ZODB changes; Kept to rebuild δBtail/byRoot after new Track.
	// includes all changed objects, not only tracked ones.
	δZtail *zodb.ΔTail

	// XXX vvv keys keys ∈ tracket -> keys ∈ kadj[tracket] ?
//	δRtail []ΔRoots // which BTree were changed; Noted only by keys ∈ tracket subset
//	byRoot map[*Tree]*ΔTtail // root -> k/v change history; only for keys ∈ tracket subset
	byRoot map[zodb.Oid]*ΔTtail // {} root -> [] k/v change history; only for keys ∈ tracket subset

	// XXX or ask client to provide db on every call?
	db     *zodb.DB    // to open connections to load new/old tree|buckets

	// tracked nodes index: node -> parent + accessed holes under this node
	// we only allow single parent/root case and report "tree corrupt" otherwise.
	// trackIdx describes @head state
	trackIdx map[zodb.Oid]nodeTrack

	// tracked objects that are not yet taken into account in current δBtail
	trackNew SetOid
}

// XXX place
// nodeTrack represents tracking information about a node.
type nodeTrack struct {
	parent  zodb.Oid // parent node | InvalidOid for root
	holes   SetKey   // missing keys tracked under this node;  nil for !leaf
}

// ΔB represents a change in BTrees space.
type ΔB struct {
	Rev	zodb.Tid
	ByRoot	map[zodb.Oid]map[Key]ΔValue // {} root -> {}(key, δvalue)
}

/*
// ΔRoots describes which BTrees were change in one revision.
type ΔRoots struct {
	Rev   zodb.Tid
	Roots []*Tree	// root XXX -> Oid?	XXX -> SetTree?
}
*/

// ΔTtail represent tail of revisional changes to one BTree.
//
// See ΔBtail documentation for details.
type ΔTtail struct {
	vδT []ΔTree // changes to tree keys; rev↑. covers keys ∈ tracked subset

	// {}k/v @tail for keys that are changed in (tail, head].
	KVAtTail map[Key]Value

	// index for LastRevOf queries
	lastRevOf map[Key]zodb.Tid // {} key -> last
}

// ΔTree describes changes to one BTree in one revision.
// XXX -> ΔT ?
type ΔTree struct {
	Rev  zodb.Tid
	KV   map[Key]Value
}

// NewΔBtail creates new empty ΔBtail object.
//
// Initial tracked set is empty.
// Initial coverage is (at₀, at₀].
//
// db will be used by ΔBtail to open database connections ... XXX
// XXX or caller provides zhead/zprev explicitly?
func NewΔBtail(at0 zodb.Tid, db *zodb.DB) *ΔBtail {
	return &ΔBtail{
		δZtail:   zodb.NewΔTail(at0),
		byRoot:   make(map[zodb.Oid]*ΔTtail),
		trackIdx: map[zodb.Oid]nodeTrack{},
		db:       db,
	}
}

// newΔTtail creates new empty ΔTtail object.
func newΔTtail() *ΔTtail {
	return &ΔTtail{
		KVAtTail:  make(map[Key]Value),
		lastRevOf: make(map[Key]zodb.Tid),
	}
}

// (tail, head] coverage
func (δBtail *ΔBtail) Head() zodb.Tid { return δBtail.δZtail.Head() }
func (δBtail *ΔBtail) Tail() zodb.Tid { return δBtail.δZtail.Tail() }

// XXX SliceByRev?
// XXX ForgetPast

// Track adds tree path to tracked set.
//
// path[0] signifies tree root.
// All path elements must be Tree except last one which must be Bucket.
//
// XXX δBtail is rebuild to also include keys corresponding to added nodes.
//
// XXX objects in path must be with .PJar().At() == .head
// XXX path -> []oid ?
//
// XXX catch cycles on add?
type TrackFlags int
const (
	// TrackMaxKey
	// - indicates that provided path is currently the rightmost arm of the tree, and
	// - requests that changes to max key of the tree to be covered by δBtail from now on.
	TrackMaxKey TrackFlags = 1 << iota

	// XXX TrackMinKey (we don't need it in WCFS)
)
func (δBtail *ΔBtail) Track(key Key, keyPresent bool, path []Node, flags TrackFlags) error {	// XXX Tree|Bucket; path[0] = root
	l := len(path)
	if l == 0 {
		panic("empty path")
	}

	// XXX assert Tree Tree ... Tree Bucket
	root := path[0].(*Tree).POid()

	pathv := []string{}
	for _, node := range path { pathv = append(pathv, vnode(node)) }
	fmt.Printf("Track [%v] %s\n", key, strings.Join(pathv, " -> "))

	parent := zodb.InvalidOid
	var track    nodeTrack
	var oldTrack bool
	for i, node := range path {
		oid := node.POid()
		// XXX check for InvalidOid  (e.g. T/B1:a with bucket not having its own oid.
		track, oldTrack = δBtail.trackIdx[oid]
		if !oldTrack {
			track = nodeTrack{parent: parent}
			if i == l-1 { // leaf
				track.holes = SetKey{}
			}
			δBtail.trackIdx[oid] = track
			// XXX .trackNew += oid
		}
		if track.parent != parent {
			// XXX -> error (e.g. due to corrupt data in ZODB)
			panicf("node %s is reachable from multiple parents: %s %s",
				oid, track.parent, parent)
		}
		parent = oid
	}

	// track is track of path[-1]  (i.e. leaf)

	// remember missing keys in track of leaf node (bucket or top-level ø tree)
	if !keyPresent {
		track.holes.Add(key)
	} else {
		if track.holes.Has(key) {
			panicf("[%v] was previously requested to be tracked as ø", key)
		}
	}


	_, ok := δBtail.byRoot[root]
	if !ok {
		δBtail.byRoot[root] = newΔTtail()
	}

	// XXX update diff	XXX here? or as separate step?
	// XXX update lastRevOf
	return nil
}

// Update updates δB with per-object level ZODB changes.
//
// Only those objects from δZ that belong to tracked set are guaranteed to be
// taken into account. In other words a tree history will assuredly include
// only those keys, that correspond to tracked subset of δZ.
//
// δZ should include all objects changed by ZODB transaction.
//
// XXX optionally accept zconnOld/zconnNew from client
func (δBtail *ΔBtail) Update(δZ *zodb.EventCommit) (_ ΔB, err error) {
	headOld := δBtail.Head()
	defer xerr.Contextf(&err, "ΔBtail update %s -> %s", headOld, δZ.Tid)

	δBtail.δZtail.Append(δZ.Tid, δZ.Changev)

	fmt.Printf("\nUpdate δZ: %v\n", δZ.Changev)
	fmt.Printf("trackIdx: %v\n", δBtail.trackIdx)

	δZTC, δtopsByRoot := δBtail.δZConnectTracked(δZ)

	δB := ΔB{Rev: δZ.Tid, ByRoot: make(map[zodb.Oid]map[Key]ΔValue)}

	// skip opening DB connections if there is no change to any tree node
	if len(δtopsByRoot) == 0 {
		return δB, nil
	}

	// open ZODB connections corresponding to "old" and "new" states
	// TODO caller should provide one of those (probably new)
	txn, ctx := transaction.New(context.TODO()) // XXX
	defer txn.Abort()
	zconnOld, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: headOld})
	if err != nil {
		return ΔB{}, err
	}
	zconnNew, err := δBtail.db.Open(ctx, &zodb.ConnOptions{At: δZ.Tid})
	if err != nil {
		return ΔB{}, err
	}

	for root, δtops := range δtopsByRoot {
		δT, err := treediff(ctx, root, δtops, δZTC, δBtail.trackIdx, zconnOld, zconnNew)
		if err != nil {
			return ΔB{}, err
		}

		if len(δT) > 0 { // an object might be resaved without change
			δB.ByRoot[root] = δT
		}
	}

	// XXX .δBtail <- (δZ.rev, [](root, []key))
	// XXX rebuild lastRevOf

	return δB, nil
}

// δZConnectTracked computes connected closure of δZ/T.
//
// δZ    - all changes in a ZODB transaction.
// δZ/T  - subset of those changes intersection with tracking set.
// δZ/TC - connected closure for δZ/T
//
// for example for e.g. t₀->t₁->b₂ if δZ/T={t₀ b₂} -> δZ/TC=δZ/T+{t₁}
//
// δtopsByRoot = {} root -> {top changed nodes in that tree}
func (δBtail *ΔBtail) δZConnectTracked(δZv *zodb.EventCommit) (δZTC SetOid, δtopsByRoot map[zodb.Oid]SetOid) {
	δZ  := SetOid{};  for _, δ := range δZv.Changev { δZ.Add(δ) }
	δZTC = SetOid{}
	δtopsByRoot = map[zodb.Oid]SetOid{}

	for δ := range δZ {
		track, ok := δBtail.trackIdx[δ]
		if !ok {
			continue // not tracked at all
		}

		δZTC.Add(δ)

		// go up by .parent till root or till another tracked node in the tree
		// if root  -> δtopsByRoot[root] += δ
		// if !root -> δZTC += path through which we reached another node (forming connection)
		path := []zodb.Oid{}
		node := δ
		parent := track.parent
		for {
			// reached root
			if parent == zodb.InvalidOid {
				root := node
				δtops, ok := δtopsByRoot[root]
				if !ok {
					δtops = SetOid{}
					δtopsByRoot[root] = δtops
				}
				δtops.Add(δ)
				break
			}

			// reached another tracked node
			if δZ.Has(parent) {
				for _, δp := range path {
					δZTC.Add(δp)
				}
				break
			}

			path = append(path, parent)
			trackUp, ok := δBtail.trackIdx[parent]
			if !ok {
				panicf("BUG: .p%s -> %s, but %s is not tracked", node, parent, parent)
			}
			node = parent
			parent = trackUp.parent
		}
	}

	return δZTC, δtopsByRoot
}


// XXX place
// nodeInRange represents a Node coming under [lo, hi_] key range in its tree.
type nodeInRange struct {
	lo, hi_  Key  // [lo, hi_]	NOTE _not_ hi) not to overflow at ∞
	node     Node
	done     bool // whether this node was already taken into account while computing diff
}

// rangeSplit represents set of nodes covering a range.
// nodes come with key↑ and no intersection in between their [lo,hi)
type rangeSplit []*nodeInRange // key↑

// Get returns node covering key k.
// Get panics if k is not covered.
func (rs rangeSplit) Get(k Key) *nodeInRange {
	rnode, ok := rs.Get_(k)
	if !ok {
		panicf("key %v not covered;  coverage: %s", k, rs)
	}
	return rnode
}

// Get_ returns node covering key k.
func (rs rangeSplit) Get_(k Key) (rnode *nodeInRange, ok bool) {
	i := sort.Search(len(rs), func(i int) bool {
		return k <= rs[i].hi_
	})
	if i == len(rs) {
		return nil, false // key not covered
	}

	rn := rs[i]
	if !(rn.lo <= k && k <= rn.hi_) {
		panicf("BUG: get(%v) -> %s;  coverage: %s", k, rn, rs)
	}

	return rn, true
}

// Expand replaces rnode with its children.
//
// rnode must be initially in *prs.
// rnode.node must be tree.
// rnode.node must be already activated.
//
// inserted children is also returned for convenience.
func (prs *rangeSplit) Expand(rnode *nodeInRange) (children rangeSplit) {
	rs := *prs
	i := sort.Search(len(rs), func(i int) bool {
		return rnode.hi_ <= rs[i].hi_
	})
	if i == len(rs) || rs[i] != rnode {
		panicf("%s not in rangeSplit;  coverage: %s", rnode, rs)
	}

	// [i].Key ≤ [i].Child.*.Key < [i+1].Key  i ∈ [0, len([]))
	//
	// [0].Key       = -∞ ; always returned so
	// [len(ev)].Key = +∞ ; should be assumed so
	tree  := rnode.node.(*Tree)
	treev := tree.Entryv()
	children = make(rangeSplit, 0, len(treev)+1)
	for i := range treev {
		lo := rnode.lo
		if i > 0 {
			lo = treev[i].Key()
		}
		hi_ := rnode.hi_
		if i < len(treev)-1 {
			hi_ = treev[i+1].Key()-1 // NOTE -1 because it is hi_] not hi)
		}

		children = append(children, &nodeInRange{
					lo:   lo,
					hi_:  hi_,
					node: treev[i].Child(),
		})
	}

	// del[i]; insert(@i, children)
	*prs = append(rs[:i], append(children, rs[i+1:]...)...)
	return children
}

// GetToLeaf returns leaf node corresponding to key k.
//
// Leaf is usually bucket node, but, in the sole single case of empty tree, can be that root tree node.
// GetToLeaf expands step-by-step every tree from where it has to traverse to next depth level.
//
// XXX also return path?
func (prs *rangeSplit) GetToLeaf(ctx context.Context, k Key) (rnode *nodeInRange, ok bool, err error) {
	rnode, ok = prs.Get_(k)
	if !ok {
		return nil, false, nil // key not covered
	}

	for {
		switch rnode.node.(type) {
		// bucket = leaf
		case *Bucket:
			return rnode, true, nil
		}

		// its tree -> activate to expand; check for ø case
		tree := rnode.node.(*Tree)
		err = tree.PActivate(ctx)
		if err != nil {
			return nil, false, err
		}
		defer tree.PDeactivate()

		// empty tree -> don't expand - it is already leaf
		if len(tree.Entryv()) == 0{
			return rnode, true, nil
		}

		// expand tree children
		children := prs.Expand(rnode)
		rnode = children.Get(k) // k must be there
	}
}

func (rs rangeSplit) String() string {
	if len(rs) == 0 {
		return "ø"
	}

	s := ""
	for _, rn := range rs {
		if s != "" {
			s += " "
		}
		s += fmt.Sprintf("%s", rn)
	}
	return s
}



// treediff computes δT for tree specified by root in between old..new.
//
// δtops is set of top nodes for changed subtrees.
// δZTC is connected(δZ/T) - connected closure for subset of δZ(old..new) that
// touches tracked nodes of T.
func treediff(ctx context.Context, root zodb.Oid, δtops SetOid, δZTC SetOid, trackIdx map[zodb.Oid]nodeTrack, zconnOld, zconnNew *zodb.Connection) (δT map[Key]ΔValue, err error) {
	defer xerr.Contextf(&err, "treediff %s..%s %s", zconnOld.At(), zconnNew.At(), root)
	// XXX zconnX -> a, b ?

	fmt.Printf("treediff %s  δtops: %v  δZTC: %v\n", root, δtops, δZTC)
	defer fmt.Printf("\n")

	δT = map[Key]ΔValue{}

	for top := range δtops { // XXX -> sorted?
		a, err1 := zgetNode(ctx, zconnOld, top)
		b, err2 := zgetNode(ctx, zconnNew, top)
		err := xerr.Merge(err1, err2)
		if err != nil {
			return nil, err
		}

		δtop, err := diffX(ctx, a, b, δZTC, trackIdx)
		if err != nil {
			return nil, err
		}

		// FIXME -> merge (VDEL vs add)
		// XXX no - not needed here - keys cannot migrate in between two disconnected subtrees
		//     -> assert that keys from different δtop do not overlap
		// DEL k -> Tkextra += k
		// +k    -> Tkextra -= k
		fmt.Printf("-> δtop: %v\n", δtop)
		for k,δv := range δtop {
			δT[k] = δv
		}
	}

	return δT, nil
}

// diffX computes difference in between two revisions of a tree's subtree.
//
// a, b point to top of the subtree @old and @new revisions and must be of the
// same type - tree or bucket.
//
// δZTC is connected set of objects covering δZT (objects changed in this tree in old..new).
//
// a/b can be nil; a=nil means addition, b=nil means deletion.
func diffX(ctx context.Context, a, b Node, δZTC SetOid, trackIdx map[zodb.Oid]nodeTrack) (δ map[Key]ΔValue, err error) {
	if a==nil && b==nil {
		panic("BUG: both a & b == nil")
	}

	var aT, bT *Tree
	var aB, bB *Bucket
	isT := false

	if a != nil {
		aT, isT = a.(*Tree)
		aB, _   = a.(*Bucket)
		if aT == nil && aB == nil {
			panicf("a: bad type %T", a)
		}
	}
	if b != nil {
		bT, isT = b.(*Tree)
		bB, _   = b.(*Bucket)
		if bT == nil && bB == nil {
			panicf("b: bad type %T", b)
		}
	}

	if a != nil && b != nil {
		if a.POid() != b.POid() {
			panicf("BUG: a.oid != b.oid  ; a: %s  b: %s", a.POid(), b.POid())
		}
		if !((aT != nil && bT != nil) || (aB != nil && bB != nil)) {
			return nil, fmt.Errorf("object %s: type mutated %s -> %s", a.POid(),
				zodb.ClassOf(a), zodb.ClassOf(b))
		}
	}

	if isT {
		return diffT(ctx, aT, bT, δZTC, trackIdx)
	} else {
		return diffB(ctx, aB, bB)
	}
}

// diffT computes difference in between two subtrees.
//
// a, b point to top of subtrees @old and @new revisions.
// δZTC is connected set of objects covering δZT (objects changed in this tree in old..new).
//
// XXX trackIdx -> just pass δBtail?
func diffT(ctx context.Context, a, b *Tree, δZTC SetOid, trackIdx map[zodb.Oid]nodeTrack) (δ map[Key]ΔValue, err error) {
	fmt.Printf("  T %s %s\n", xidOf(a), xidOf(b))
	defer xerr.Contextf(&err, "diffT %s %s", xidOf(a), xidOf(b))

	δ = map[Key]ΔValue{}
	defer fmt.Printf("    -> δt: %v\n", δ)

	var av rangeSplit  // nodes expanded from a
	var bv rangeSplit  // nodes expanded from b
	Aqueue := SetKey{} // "to process" keys on A
	Bqueue := SetKey{} // "to process" keys on B
	Adone  := SetKey{} // "processed" keys on A
	Bdone  := SetKey{} // "processed" keys on B

	if b != nil { // XXX kill (always !nil) ?
		// XXX precise range as for a ^^^ ?
		btop := &nodeInRange{lo: KeyMin, hi_: KeyMax, node: b} // [-∞, ∞)
		bv   = rangeSplit{btop}
	}

	// initial phase: expand changed nodes in a till buckets;
	// XXX changed buckets -> δ-
	if a != nil {	// XXX kill (always !nil) ?
		// XXX maybe walk till a from root to get more precise initial range?
		atop := &nodeInRange{lo: KeyMin, hi_: KeyMax, node: a} // [-∞, ∞)
		av  = rangeSplit{atop}
		fmt.Printf("X av: %s\n", av)
		aq := []*nodeInRange{atop} // stack
		for len(aq) > 0 {
			l := len(aq)
			arn  := aq[l-1]; aq=aq[:l-1] // arn=aq.pop()
			atree := arn.node.(*Tree)    // ok - only trees in aq
			err = atree.PActivate(ctx);  if err != nil { return nil, err}
			defer atree.PDeactivate()

			if len(atree.Entryv()) == 0 {
				// empty tree - do not expand into bucket - only process tracked holes
				// XXX dup wrt bucket processing?
				track, ok := trackIdx[atree.POid()]
				if !ok {
					panicf("%s ∈ δZTC, but ∉ trackIdx", vnode(atree))
				}
				δA := map[Key]ΔValue{}
				for k := range track.holes {
					δA[k] = ΔValue{VDEL, VDEL} // ø->ø indicates hole
				}
				// δ <- δA
				err = δMerge(δ, δA)
				if err != nil {
					return nil, err
				}

				// Adone  <- δA
				// Bqueue <- δA
				for k := range δA {
					Adone.Add(k)
					Bqueue.Add(k)
				}

				arn.done = true
				continue
			}

			children := av.Expand(arn)
			for _, rchild := range children {
				coid := rchild.node.POid()
				if !( δZTC.Has(coid) ||
				      /* embedded bucket */
				      (len(children) == 1 && coid == zodb.InvalidOid) )  {
					continue
				}

				switch node := rchild.node.(type) {
				case *Tree:
					aq = append(aq, rchild)

				case *Bucket:
					δA, err := diffB(ctx, node, nil)
					if err != nil {
						return nil, err
					}
					// also -[k]ø (for tracked holes)
					track, ok := trackIdx[node.POid()]
					if !ok {
						panicf("%s ∈ δZTC, but ∉ trackIdx", vnode(node))
					}
					for k := range track.holes {
						δA[k] = ΔValue{VDEL, VDEL} // ø->ø indicates hole
					}

					// δ <- δA
					err = δMerge(δ, δA)
					if err != nil {
						return nil, err
					}

					// Adone  <- δA
					// Bqueue <- δA
					for k := range δA {
						Adone.Add(k)
						Bqueue.Add(k)
					}

					rchild.done = true
				}
			}
		}
	}

	fmt.Printf("  av: %s\n", av)
	fmt.Printf("  bv: %s\n", bv)
	fmt.Printf("  bq: %s\n", Bqueue)

	// phase 2: iterate through keys queued for A and B, delve into
	// corresponding nodes, and merge diff generated from them into δ.
	// Each delve for A or B, potentially adds new keys to process on the
	// other side.
	//
	// XXX inefficient: we process each key separately, while they can be
	// processed in sorted batches.
	for len(Aqueue) > 0 || len(Bqueue) > 0 {
		fmt.Printf("\n")
		fmt.Printf("  aq: %s\n", Aqueue)
		fmt.Printf("  bq: %s\n", Bqueue)

		// B queue
		// expand keys in new δA -> in B till buckets;
		// process B buckets that cover new keys into δ+
		Aqueue = SetKey{}
		for k := range Bqueue {
			fmt.Printf("  B [%v]\n", k)
			bnode, ok, err := bv.GetToLeaf(ctx, k)
			if err != nil {
				return nil, err
			}
			if !ok {
				// FIXME -> key must be included into some node.hole
				continue // key not covered
			}

			// +bucket if that bucket is reached for the first time
			if !bnode.done {
				var δB map[Key]ΔValue
				bbucket, ok := bnode.node.(*Bucket)
				if ok { // !ok means ø tree
					δB, err = diffB(ctx, nil, bbucket)
					if err != nil {
						return nil, err
					}
				}

				// δ <- δB
				err = δMerge(δ, δB)
				if err != nil {
					return nil, err
				}

				// Bdone  <- δB
				// Aqueue <- δB
				for k_ := range δB {
					Bdone.Add(k_)
					if !Adone.Has(k_) {
						Aqueue.Add(k_)
					}
				}

				bnode.done = true
			}

			// k is not there -> +[k]ø
			if !Bdone.Has(k) {
				δB := map[Key]ΔValue{k: {VDEL, VDEL}} // [k]ø->ø
				err = δMerge(δ, δB);  if err != nil { return nil, err }
				Bdone.Add(k)
			}

			fmt.Printf("  bv: %s\n", bv)
		}

		// A queue
		Bqueue = SetKey{}
		for k := range Aqueue {
			fmt.Printf("  A [%v]\n", k)
			anode, ok, err := av.GetToLeaf(ctx, k)
			if err != nil {
				return nil, err
			}
			if !ok {
				// FIXME -> key must be included into some node.hole
				continue // key not covered
			}

			// XXX check for anode.node.(*Tree)	(ø tree case)

			// - bucket if that bucket is reached for the first time
			if !anode.done {
				δA, err := diffB(ctx, anode.node.(*Bucket), nil)
				if err != nil {
					return nil, err
				}
				// XXX also extract holes

				// δ <- δA
				err = δMerge(δ, δA)
				if err != nil {
					return nil, err
				}

				// Adone  <- δA
				// Bqueue <- δA
				for k_ := range δA {
					Adone.Add(k_)
					if !Bdone.Has(k_) {
						Bqueue.Add(k_)
					}
				}

				anode.done = true
			}

			// k is not there -> -[k]ø
			if !Adone.Has(k) {
				// XXX do we need to add(?) [k]ø->ø ?
				Adone.Add(k)
			}

			fmt.Printf("  av: %s\n", av)
		}
	}

	return δ, nil
}


// δMerge merges changes from δ2 into δ.
// δ is total-building diff, while δ2 is diff from comparing some subnodes.
func δMerge(δ, δ2 map[Key]ΔValue) error {
	fmt.Printf("  δmerge %v <- %v\n", δ, δ2)
	defer fmt.Printf("      -> %v\n", δ)

	// merge δ <- δ2
	for k, δv2 := range δ2 {
		δv1, already := δ[k]
		if !already {
			δ[k] = δv2
			continue
		}

		// both δ and δ2 has [k] - it can be that key
		// entry migrated from one bucket into another.
		// XXX also handle ø->ø + ø->c  i.e. [k] was hole and become value
		if !( (δv1.New == VDEL && δv2.Old == VDEL) ||
		      (δv1.Old == VDEL && δv2.New == VDEL) ) {
			return fmt.Errorf("BUG or btree corrupt: [%v] has " +
					  "duplicate entries: %v, %v", k, δv1, δv2)
		}

		δv := ΔValue{}
		switch {

		// x -> ø | ø -> x
		// ø -> ø
		case δv2.Old == VDEL && δv2.New == VDEL: // δv2 == hole
			δv = δv1

		// ø -> ø
		// y -> ø | ø -> y
		case δv1.Old == VDEL && δv1.New == VDEL: // δv1 == hole
			δv = δv2

		// ø -> x		-> y->x
		// y -> ø
		case δv2.New == VDEL:
			δv.Old = δv2.Old
			δv.New = δv1.New

		// x -> ø		-> x->y
		// ø -> y
		default:
			δv.Old = δv1.Old
			δv.New = δv2.New
		}

		fmt.Printf("      [%v] merge %s %s  -> %s\n", k, δv1, δv2, δv)
		if δv.Old != δv.New {
			δ[k] = δv
		} else {
			delete(δ, k) // NOTE also annihilates hole migration
		}
	}

	return nil
}



/*
func __diffT(ctx context.Context, a, b *Tree, δZTC SetOid, trackIdx map[zodb.Oid]nodeTrack) (δ map[Key]ΔValue, err error) {
	fmt.Printf("  T %s %s\n", xidOf(a), xidOf(b))
	defer xerr.Contextf(&err, "diffT %s %s", xidOf(a), xidOf(b))

	var av []TreeEntry
	var bv []TreeEntry
	if a != nil {
		err = a.PActivate(ctx);  if err != nil { return nil, err }
		defer a.PDeactivate()
		av  = a.Entryv() // key↑
	}
	if b != nil {
		err = b.PActivate(ctx);  if err != nil { return nil, err }
		defer b.PDeactivate()
		bv  = b.Entryv() // key↑
	}
	fmt.Printf("  av: %s\n", vtree(av))
	fmt.Printf("  bv: %s\n", vtree(bv))

	δ = map[Key]ΔValue{}
	defer fmt.Printf("    -> δt: %v\n", δ)

	// [i].Key ≤ [i].Child.*.Key < [i+1].Key  i ∈ [0, len([]))
	//
	// [0].Key       = -∞ ; always returned so
	// [len(ev)].Key = +∞ ; should be assumed so

	// a -> {aChildren} : ∈ δZTC
	aChildren := map[zodb.Oid]Node{}
	for _, __ := range av {
		child := __.Child()
		coid  := child.POid()
		if δZTC.Has(coid) ||
		   (len(av) == 1 && coid == zodb.InvalidOid) { // embedded bucket {
			aChildren[coid] = child
		}
	}
	fmt.Printf("  aChildren: %v\n", aChildren)

	// -> "might be changed" subset of tracked keys
	δtqkeys := SetKey{}
	for achild := range aChildren {
		track, ok := trackIdx[achild]
		if !ok {
			panicf("node %s ∈ δZTC, but its track is ø", achild)
		}
		δtqkeys.Update(track.trackedKeys)
	}
	if a != nil && len(av) == 0 { // ø tree - fetch tracked keys from it directly
		δtqkeys.Update(trackIdx[a.POid()].trackedKeys)
	}
	fmt.Printf("  δtqkeys:   %s\n", δtqkeys)

	// b -> {bChildren} : coverage ∩ "might be changed" tracked keys
	bChildren := map[zodb.Oid]Node{}
	for i := range bv {
		lo  := bv[i].Key()  // [lo, hi_]   _not_ hi) not to overflow at ∞
		hi_ := KeyMax
		if i+1 < len(bv) {
			hi_ = bv[i+1].Key() - 1
		}
		child := bv[i].Child()

		// XXX dumb
		for k := range δtqkeys {
			if lo <= k && k <= hi_ {
				bChildren[child.POid()] = child // InvalidOid = embedded bucket
				break
			}
		}
	}
	fmt.Printf("  bChildren: %v\n", bChildren)

	// allChildren = aChildren ∪ bChildren
	allChildren := SetOid{}
	for c := range aChildren { allChildren.Add(c) }
	for c := range bChildren { allChildren.Add(c) }

	// go through all children that potentially can add to δ and process add/del/modify
	for child := range allChildren { // XXX -> sorted? FIXME wrong to traverse by oid - need to ...?
		ca := aChildren[child]
		cb := bChildren[child]

		δc, err := diffX(ctx, ca, cb, δZTC, trackIdx)
		if err != nil {
			return nil, err
		}

		// merge δ <- δc
		for k, δv1 := range δc {
			δv2, already := δ[k]
			if !already {
				δ[k] = δv1
				continue
			}

			// both δ and δc has [k] - it can be that key
			// entry migrated from one bucket into another.
			if !( (δv2.New == VDEL && δv1.Old == VDEL) ||
			      (δv2.Old == VDEL && δv1.New == VDEL) ) {
				return nil, fmt.Errorf("BUG or btree corrupt: [%v] has " +
						"duplicate entries: %v, %v", k, δv1, δv2)
			}

			δv := ΔValue{}
			if δv1.New == VDEL {
				δv.Old = δv1.Old
				δv.New = δv2.New
			} else {
				δv.Old = δv2.Old
				δv.New = δv1.New
			}

			fmt.Printf("    [%v] merge %s %s  -> %s\n", k, δv1, δv2, δv)
			if δv.Old != δv.New {
				δ[k] = δv
			} else {
				delete(δ, k)
			}
		}

		// update δc -> tracked keys

		// ca = nil -> add    cb to   tracked
		// cb = nil -> remove ca from tracked
		// (their siblings must be already processed by diffX call)
		if ca == nil {
			trackIdx[child] = nodeTrack{parent: b.POid(), trackedKeys: SetKey{}}
		}
		if cb == nil {
			delete(trackIdx, child) // XXX remove keys from parent?
		} else {
			trackedKeys := trackIdx[child].trackedKeys
			for k, δv := range δc {
				switch {
				case δv.Old == VDEL:
					trackedKeys.Add(k)

				case δv.New == VDEL:
					trackedKeys.Add(k)

				// k v1->v2   no change in key
				}
			}
		}


	}
*/

/*
	allChildren := SetOid{}
	for _, __ := range av {
		child := __.Child()
		aChildren[child.POid()] = child // InvalidOid = embedded bucket
		allChildren.Add(child.POid())
	}
	for _, __ := range bv {
		child := __.Child()
		bChildren[child.POid()] = child // ----//----
		allChildren.Add(child.POid())
	}

	for child := range allChildren { // XXX -> sorted?
		// add/del/modify, but child is not tracked - ignore
		// XXX correct only in children from a? (b could have changed children to new objects)
		if !δZTC.Has(child) {
			continue
		}

		ca := aChildren[child]
		cb := bChildren[child]

		δc, err := diffX(ctx, ca, cb, δZTC)
		if err != nil {
			return nil, err
		}

		// merge δ <- δc
		for k, δv1 := range δc {
			δv2, already := δ[k]
			if !already {
				δ[k] = δv1
				continue
			}

			// both δ and δc has [k] - it can be that key
			// entry migrated from one bucket into another.
			if !( (δv2.New == VDEL && δv1.Old == VDEL) ||
			      (δv2.Old == VDEL && δv1.New == VDEL) ) {
				return nil, fmt.Errorf("BUG or btree corrupt: [%v] has " +
						"duplicate entries: %v, %v", k, δv1, δv2)
			}

			δv := ΔValue{}
			if δv1.New == VDEL {
				δv.Old = δv1.Old
				δv.New = δv2.New
			} else {
				δv.Old = δv2.Old
				δv.New = δv1.New
			}

			if δv.Old != δv.New {
				δ[k] = δv
			} else {
				delete(δ, k)
			}
		}

/// *
		// merge Tkdel <- δc
		for k, v := range δc {
			if v == VDEL {
				Tkdel.Add(k)
			} else {
				delete(Tkdel, k)
			}
		}
//* /

		// XXX process keys from δ outside of already tracked nodes
		// XXX only deleted keys - no - delete can be from untracked
		// node and on tracked it is seen as addition
	}
*/

/*
	return δ, nil
}
*/

// diffB computes difference in between two buckets.
// see diffX for details.
func diffB(ctx context.Context, a, b *Bucket) (δ map[Key]ΔValue, err error) {
	fmt.Printf("  B %s %s\n", xidOf(a), xidOf(b))
	defer xerr.Contextf(&err, "diffB %s %s", xidOf(a), xidOf(b))
	// XXX oid can be InvalidOid for T/B... (i.e. B is part of T and is not yet committed separately)

	var av []BucketEntry
	var bv []BucketEntry
	if a != nil {
		err = a.PActivate(ctx);  if err != nil { return nil, err }
		defer a.PDeactivate()
		av  = a.Entryv() // key↑
	}
	if b != nil {
		err = b.PActivate(ctx);  if err != nil { return nil, err }
		defer b.PDeactivate()
		bv  = b.Entryv() // key↑
	}

	//fmt.Println("> diffB", a.POid())
	//defer fmt.Println("< diffB")

	δ   = map[Key]ΔValue{}
	defer fmt.Printf("    -> δb: %v\n", δ)
	//fmt.Println("av:", av)
	//fmt.Println("bv:", bv)

	for len(av) > 0 || len(bv) > 0 {
		ka, va := KeyMax, VDEL
		kb, vb := KeyMax, VDEL

		if len(av) > 0 {
			ka = av[0].Key()
			va, err = vOid(av[0].Value())
			if err != nil {
				return nil, fmt.Errorf("a[%v]: %s", ka, err)
			}
		}
		if len(bv) > 0 {
			kb = bv[0].Key()
			vb, err = vOid(bv[0].Value())
			if err != nil {
				return nil, fmt.Errorf("b[%v]: %s", kb, err)
			}
		}

		switch {
		case ka < kb: // -a[0]
			δ[ka] = ΔValue{va, VDEL}
			av = av[1:]

		case ka > kb: // +b[0]
			δ[kb] = ΔValue{VDEL, vb}
			bv = bv[1:]

		// ka == kb   // va->vb
		default:
			if va != vb {
				δ[ka] = ΔValue{va, vb}
			}
			av = av[1:]
			bv = bv[1:]
		}
	}

	return δ, nil
}

// zgetNode returns btree node corresponding to zconn.Get(oid) .
func zgetNode(ctx context.Context, zconn *zodb.Connection, oid zodb.Oid) (_ Node, err error) {
	defer xerr.Contextf(&err, "getnode %s@%s", oid, zconn.At())
	xnode, err := zconn.Get(ctx, oid)
	if err != nil {
		return nil, err
	}

	node, ok := xnode.(Node)
	if !ok {
		return nil, fmt.Errorf("unexpected type: %s", zodb.ClassOf(xnode))
	}

	return node, nil
}

// vOid returns OID of a value object.
// it is an error if value is not persistent object.
func vOid(xvalue interface{}) (zodb.Oid, error) {
	value, ok := xvalue.(zodb.IPersistent)
	if !ok {
		return zodb.InvalidOid, fmt.Errorf("%T is not a persitent object", xvalue)
	}
	return value.POid(), nil
}

// xidOf return string representation of object xid.
func xidOf(obj zodb.IPersistent) string {
	if obj == nil || reflect.ValueOf(obj).IsNil() {
		return "ø"
	}
	xid := zodb.Xid{At: obj.PJar().At(), Oid: obj.POid()}
	return xid.String()
}



func (δBtail *ΔBtail) ForgetPast(revCut zodb.Tid) {
	δBtail.δZtail.ForgetPast(revCut) // XXX stub
}


// update brings .δBtail up to date by recomputing diff XXX and taking new
// entries in .δZtail into account.
// func (δBtail *ΔBtail) update()

// Get returns root[key] as of @at database state plus revision that changed it.
//
// if revExact=False - rev is upper estimate for the revision.
//
// XXX at must ∈ (tail, head]	XXX [tail ?
// XXX key must be tracked
func (δΒtail *ΔBtail) Get(ctx context.Context, root *Tree, key Key, at zodb.Tid) (value Value, ok bool, rev zodb.Tid, revExact bool, err error) {
	defer xerr.Contextf(&err, "δBtail: root<%s>: get %d @%s", root.POid(), key, at)
	// XXX key not tracked		-> panic
	// XXX at not ∈ (tail, head]	-> panic

	// XXX handle deletion

	// FIXME stub -> that only ZBlk.rev is used
	//return @head, rev=.Tail(), revExact=false

	// XXX dirty -> rebuild

	// XXX -> index lastXXXOf(key) | linear scan ↓ looking for change <= at
	δTtail := δΒtail.byRoot[root.POid()]
	if δTtail == nil {
		panicf("δBtail: root<%s> not tracked", root.POid())
	}

	for i := len(δTtail.vδT)-1; i >= 0; i-- {
		δT := δTtail.vδT[i]
		if at < δT.Rev {
			continue
		}
		value, ok = δT.KV[key]
		if ok {
			rev = δT.Rev
			revExact = true
			break
		}
	}

	// key was found in δT ∈ δTtail
	if ok {
		return
	}

	// key not in history tail.
	// either use @tail[key], if it is present, or @head[key]
	rev = δΒtail.Tail()
	revExact = false
	value, ok = δTtail.KVAtTail[key]
	if ok {
		return
	}
	// @tail[key] is not present - key was not changing in (tail, head].
	// since at ∈ (tail, head] we can use @head[key] as the result
	xvalue, ok, err := root.Get(ctx, key)
	if err != nil || !ok {
		return
	}
	v, ok := xvalue.(zodb.IPersistent)
	if !ok {
		err = fmt.Errorf("got %T; expect ZODB persistent reference", xvalue)
		return
	}

	value = v.POid()
	ok = true

	return
}

// XXX
func (δBtail *ΔBtail) SliceByRev(lo, hi zodb.Tid) /*readonly*/ []ΔB {
	δassertSlice(δBtail, lo, hi)

	panic("TODO")
}

// SliceByRootRev returns history of a tree changes in (lo, hi] range.
//
// it must be called with the following condition:
//
//	tail ≤ lo ≤ hi ≤ head
//
// the caller must not modify returned slice.
//
// Note: contrary to regular go slicing, low is exclusive while high is inclusive.
func (δBtail *ΔBtail) SliceByRootRev(root *Tree, lo, hi zodb.Tid) /*readonly*/[]ΔTree {
	δassertSlice(δBtail, lo, hi)

	panic("TODO")
}


// ---- misc ----

// δassertSlice asserts that δ.tail ≤ lo ≤ hi ≤ δ.head
func δassertSlice(δ interface { Head() zodb.Tid; Tail() zodb.Tid }, lo, hi zodb.Tid) {
	tail := δ.Tail()
	head := δ.Head()
	if !(tail <= lo && lo <= hi && hi <= head) {
		panicf("invalid slice: (%s, %s];  (tail, head] = (%s, %s]", lo, hi, tail, head)
	}
}


/*
// vtree returns human-readable representation of Tree data.
func vtree(v []TreeEntry) string {
	s := ""
	for _, e := range v {
		if s != "" {
			s += " "
		}
		s += fmt.Sprintf("%v %s", e.Key(), vnode(e.Child()))
	}
	return s
}
*/

// vnode returns brief human-readable representation of node.
func vnode(node Node) string {
	kind := "?"
	switch node.(type) {
	case *Tree:   kind = "T"
	case *Bucket: kind = "B"
	}
	return fmt.Sprintf("%s%s", kind, node.POid())
}

func (rn nodeInRange) String() string {
	slo  := "-∞";  if rn.lo  > KeyMin { slo = fmt.Sprintf("%v", rn.lo) }
	shi  := "∞";   if rn.hi_ < KeyMax { shi = fmt.Sprintf("%v", rn.hi_+1) }
	done := " ";   if rn.done         { done = "*" }
	return fmt.Sprintf("%s[%s,%s)%s", done, slo, shi, vnode(rn.node))
}

func (track nodeTrack) String() string {
	return fmt.Sprintf("{p%s h%s}", track.parent, track.holes)
}
